clear all
close all

%test base 'e' and base '2' versions of the shift-add algorithm 
% shown at http://www.quinapalus.com/efunc.html
% and compare with matlab versions.  
% I should also look at the BKM for full imaginary exponential calcs

% further references: 
% The book by Jrg Arndt is a free download from several sources;
% search for "matters computational". Chapter 33 is "Computing the
% elementary functions with limited resources", and it deals with the
% shift/add exp and log functions as well as CORDIC. It propounds the same
% exp and log algorithm as found in the Revoland and Yakoubsohn paper you
% just linked.
% 
% Both of them use paper and book Elementary Functions by Jean-Michel
% Muller as their primary source. I highly recommend the book, which is
% thorough (theorems even!) but also has an interesting explanation using a
% pair of scales and filed weights to directly determine the log or
% exponential of a weight placed on one pan. The convergence domain is
% limited, but easily met by working in log and exp base 2.


%build the look up table
%nice_num_exp
exp_lut.exp = -[1:32];
%exp(k) aka nice_num
exp_lut.nice_num = 1 + 2.^exp_lut.exp;
%k      
exp_lut.k = log2(exp_lut.nice_num);

keep_input = pi/1000
input = keep_input

%% integer 2^x function

%compute the reference value
ref = 2^input;

%split into integer and fractional parts
s = input < 0       %1 if the input value is negative
input = abs(input);
int = floor(input)  %extract integer part
fractional = input - int  %extract fractional part

%deal with integer part
y_int = 1.0
if int > 127    %the result is larger than a single precision number can hold
    y_int = NaN
end
while int > 30  %prepare the integer part in sections because we only have 32-bit ints. 
    y_int = y_int*2^30
    int = int - 30
end
y_int = y_int*2^int         %simulate a left shift with the integer part.

%deal with fractional part
y = 1
frac = fractional
for idx = 1:32
    temp = exp_lut.k(idx)
    if temp < frac
        frac = frac - temp
        y = y + y*2^exp_lut.exp(idx)
    end
end
%combine the portions of the answer
y_out = y * y_int
if s        %correct for the input sign
    y_out = 1/y_out
end

%evaluate the results
ref
residual = frac
rel_error = (y_out-ref)/ref



%% write 2^x again with the floats fully unpacked into s, x, and m
% variables
input = keep_input


%unpack
s = input < 0       %1 if the input value is negative
input = abs(input);
x = floor(log2(input))
m = input / 2^x     %m is a binary fraction that is always between 1 and 2

%deal with integer part
xx = floor(input)    %integer part of input is the exponent of the output
                    %I can use Ftrunc for this part
mm = input - xx     %extract fractional part.  
                    
%deal with fractional part
yy = 1
%mm = m-1 %remove leading one
for idx = 1:32
    temp = exp_lut.k(idx)
    if temp < mm
        mm = mm - temp
        yy = yy + yy*2^exp_lut.exp(idx)
    end
end

mmm = yy

%pack
if s == 0
    output = mmm * 2^xx * (1 - s*2)
else
    output = 1/(mmm * 2^xx)
end

%% calculate log2()
% variables
input = keep_input
ref_log = log2(input)

%unpack
s = input < 0       %1 if the input value is negative
input = abs(input);
x = floor(log2(input))% exponent of input, also integer part of output
m = input / 2^x     %m is a binary fraction that is always between 1 and 2

mm = m/2    %make mantissa less than one
xx = x+1    %adjust integer portion of output to compensate for the division

y = 0
for idx = 1:32
    temp = mm + mm * 2^exp_lut.exp(idx)
    if temp < 1
        mm = temp
        y = y - exp_lut.k(idx)
    end
end

output = xx+y

%this adds 1.7 bits with an integer add and subtract.  worth using.
output_resid = xx + (y + (mm-1))

resid_improvement = (output-ref_log)/(output_resid - ref_log)










